from libcpp cimport bool


cdef extern from "coin/IpSmartPtr.hpp" namespace "Ipopt":
    cdef cppclass SmartPtr[T]:
        SmartPtr()
        SmartPtr(T*)
        T& operator*()
        SmartPtr[T]& operator=(T*)

    bool IsNull[U](const SmartPtr[U]&)



from libcpp cimport bool
from libcpp.string cimport string
from cython.operator cimport dereference as d
cimport pypopt.ipopt as ip


cdef class Journalist:
    cdef ip.SmartPtr[ip.Journalist] c_journalist

    def printf(self, level, category, fmt, *args, **kwargs):
        py_bytes = fmt.format(*args, **kwargs).encode('utf-8')
        cdef const char *c_str = py_bytes
        cdef ip.EJournalCategory c_category = category
        cdef ip.EJournalLevel c_level = level
        d(self.c_journalist).Printf(c_level, c_category, c_str)

    def print_string_over_lines(self, level, category, ip.Index indent_level,
                                ip.Index max_length, str line):
        cdef string c_line = line.encode('utf-8')
        cdef ip.EJournalCategory c_category = category
        cdef ip.EJournalLevel c_level = level
        d(self.c_journalist).PrintStringOverLines(
            c_level, c_category, indent_level, max_length, c_line,
        )

    def printf_indented(self, level, category, ip.Index indent_level, fmt, *args, **kwargs):
        py_bytes = fmt.format(*args, **kwargs).encode('utf-8')
        cdef const char *c_str = py_bytes
        cdef ip.EJournalCategory c_category = category
        cdef ip.EJournalLevel c_level = level
        d(self.c_journalist).PrintfIndented(c_level, c_category, indent_level, c_str)

    def can_produce_output(self, level, category):
        cdef ip.EJournalCategory c_category = category
        cdef ip.EJournalLevel c_level = level
        return d(self.c_journalist).ProduceOutput(c_level, c_category)

    def flush_buffer(self):
        d(self.c_journalist).FlushBuffer()

    def add_journal(self, Journal journal):
        cdef ip.SmartPtr[ip.Journal] c_journal = journal.c_journal
        return d(self.c_journalist).AddJournal(c_journal)

    def add_file_journal(self, str location_name, str fname, default_level=None):
        cdef ip.EJournalLevel c_level = ip.EJournalLevel.J_WARNING
        if default_level is not None:
            c_level = default_level
        cdef string c_location_name = location_name.encode('utf-8')
        cdef string c_fname = fname.encode('utf-8')
        c_journal = d(self.c_journalist).AddFileJournal(c_location_name, c_fname, c_level)
        if ip.IsNull(c_journal):
            return None
        journal = Journal()
        journal.c_journal = c_journal
        return journal

    def get_journal(self, str name):
        cdef string c_name = name.encode('utf-8')
        c_journal = d(self.c_journalist).GetJournal(c_name)
        if ip.IsNull(c_journal):
            return None
        journal = Journal()
        journal.c_journal = c_journal
        return journal

    def delete_all_journals(self):
        d(self.c_journalist).DeleteAllJournals()


cpdef enum EJournalLevel:
    J_INSUPPRESSIBLE=-1,
    J_NONE=0,
    J_ERROR,
    J_STRONGWARNING,
    J_SUMMARY,
    J_WARNING,
    J_ITERSUMMARY,
    J_DETAILED,
    J_MOREDETAILED,
    J_VECTOR,
    J_MOREVECTOR,
    J_MATRIX,
    J_MOREMATRIX,
    J_ALL,
    J_LAST_LEVEL


cpdef enum EJournalCategory:
    J_DBG=0,
    J_STATISTICS,
    J_MAIN,
    J_INITIALIZATION,
    J_BARRIER_UPDATE,
    J_SOLVE_PD_SYSTEM,
    J_FRAC_TO_BOUND,
    J_LINEAR_ALGEBRA,
    J_LINE_SEARCH,
    J_HESSIAN_APPROXIMATION,
    J_SOLUTION,
    J_DOCUMENTATION,
    J_NLP,
    J_TIMING_STATISTICS,
    J_USER_APPLICATION  ,
    J_USER1  ,
    J_USER2  ,
    J_USER3  ,
    J_USER4  ,
    J_USER5  ,
    J_USER6  ,
    J_USER7  ,
    J_USER8  ,
    J_USER9  ,
    J_USER10  ,
    J_USER11  ,
    J_USER12  ,
    J_USER13  ,
    J_USER14  ,
    J_USER15  ,
    J_USER16  ,
    J_USER17  ,
    J_LAST_CATEGORY



import pytest
from tests.conftest import app


def test_can_set_string_options(app):
    assert app.options().set_string_value('output_file', '/tmp/test.file')
    assert app.options().get_string_value('output_file') == '/tmp/test.file'


def test_can_set_numeric_options(app):
    assert app.options().set_numeric_value('tol', 1e-6)
    assert app.options().get_numeric_value('tol') == 1e-6



from libcpp cimport bool
from libcpp.string cimport string
from pypopt.coin.return_codes cimport *
from pypopt.coin.options_list cimport *
from pypopt.coin.smart_ptr cimport *
from pypopt.coin.journalist cimport *
from pypopt.coin.solve_statistics cimport *
from pypopt.coin.tnlp cimport *


cdef extern from "coin/IpIpoptApplication.hpp" namespace "Ipopt":
    cdef cppclass IpoptApplication:
        IpoptApplication() except +
        ApplicationReturnStatus Initialize(string, bool)
        ApplicationReturnStatus Initialize(bool)
        ApplicationReturnStatus OptimizeTNLP(const SmartPtr[TNLP]&)
        bool RethrowNonIpoptException(bool)
        void PrintCopyrightMessage()
        SmartPtr[OptionsList] Options()
        SmartPtr[Journalist] Jnlst()
        SmartPtr[SolveStatistics] Statistics()



cdef extern from "coin/IpReturnCodes.hpp" namespace "Ipopt":
    cdef enum ApplicationReturnStatus:
        Solve_Succeeded = 0
        Solved_To_Acceptable_Level = 1
        Infeasible_Problem_Detected = 2
        Search_Direction_Becomes_Too_Small = 3
        Diverging_Iterates = 4
        User_Requested_Stop = 5
        Feasible_Point_Found = 6
        Maximum_Iterations_Exceeded = -1
        Restoration_Failed = -2
        Error_In_Step_Computation = -3
        Maximum_CpuTime_Exceeded = -4
        Not_Enough_Degrees_Of_Freedom = -10
        Invalid_Problem_Definition = -11
        Invalid_Option = -12
        Invalid_Number_Detected = -13
        Unrecoverable_Exception = -100
        NonIpopt_Exception_Thrown = -101
        Insufficient_Memory = -102
        Internal_Error = -199

    cdef enum AlgorithmMode:
        RegularMode
        RestorationPhaseMode



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cpdef enum IndexStyle:
    C_STYLE
    FORTRAN_STYLE


cdef class NLPInfo:
    cdef readonly ip.Index n, m, nnz_jac, nnz_hess
    cdef readonly ip.IndexStyleEnum index_style

    def __init__(self, n, m, nnz_jac, nnz_hess, index_style=None):
        cdef ip.IndexStyleEnum c_index_style = ip.IndexStyleEnum.C_STYLE
        self.n = n
        self.m = m
        self.nnz_jac = nnz_jac
        self.nnz_hess = nnz_hess
        if index_style is not None:
            c_index_style = index_style
        self.index_style = c_index_style


cdef class TNLP:
    cpdef get_nlp_info(self):
        """Return information about the problem."""
        pass

    cpdef get_bounds_info(self, x_l, x_u, g_l, g_u):
        """Return bounds information about variables and constraints."""
        pass

    cpdef get_starting_point(self, init_x, x, init_z, z_l, z_u, init_lambda, lambda_):
        """Return the starting point."""
        pass

    cpdef eval_f(self, x, new_x):
        """Return the value of the objective function."""
        pass

    cpdef eval_grad_f(self, x, new_x, grad_f):
        """Return the value of the gradient of the objective function."""
        pass

    cpdef eval_g(self, x, new_x, g):
        """Return the value of the constraints."""
        pass

    cpdef get_jac_g_structure(self, row, col):
        """Return the (sparse) structure of the jacobian."""
        pass

    cpdef eval_jac_g(self, x, new_x, values):
        """Return the value of the jacobian."""
        pass

    cpdef get_h_structure(self, row, col):
        """Return the (sparse) structure of the hessian."""
        pass

    cpdef eval_h(self, x, new_x, obj_factor, lambda_, new_lambda, values):
        """Return the value of the hessian."""
        pass

    cpdef finalize_solution(self, status, x, z_l, z_u, g, lambda_, obj_value):
        """This method is called when the algorithm finishes."""
        pass



from libcpp cimport bool
from pypopt.coin.alg_types cimport *
from pypopt.coin.typedef cimport *
from pypopt.coin.return_codes cimport *
from pypopt.coin.ipopt_data cimport *
from pypopt.coin.ipopt_calculated_quantities cimport *


cdef extern from "coin/IpTNLP.hpp" namespace "Ipopt":
    ctypedef enum IndexStyleEnum "Ipopt::TNLP::IndexStyleEnum":
        C_STYLE "Ipopt::TNLP::C_STYLE"
        FORTRAN_STYLE "Ipopt::TNLP::FORTRAN_STYLE"

    cdef cppclass TNLP:
        TNLP()
        void finalize_solution(SolverReturn, Index, const Number *, const Number *, const Number *,
                               Index, const Number *, const Number *, Number, const IpoptData *,
                               IpoptCalculatedQuantities *)

        bool intermediate_callback(AlgorithmMode, Index, Number, Number, Number, Number, Number,
                                   Number, Number, Number, Index, const IpoptData *,
                                   IpoptCalculatedQuantities *)



cimport ipopt as ip

cpdef enum SolverReturn:
    SUCCESS = ip.SUCCESS
    MAXITER_EXCEEDED
    CPUTIME_EXCEEDED
    STOP_AT_TINY_STEP
    STOP_AT_ACCEPTABLE_POINT
    LOCAL_INFEASIBILITY
    USER_REQUESTED_STOP
    FEASIBLE_POINT_FOUND
    DIVERGING_ITERATES
    RESTORATION_FAILURE
    ERROR_IN_STEP_COMPUTATION
    INVALID_NUMBER_DETECTED
    TOO_FEW_DEGREES_OF_FREEDOM
    INVALID_OPTION
    OUT_OF_MEMORY
    INTERNAL_ERROR
    UNASSIGNED



cdef class ApplicationReturnStatus:
    cdef ip.ApplicationReturnStatus c_status

    def __cinit__(self, ip.ApplicationReturnStatus status):
        self.c_status = status

    def is_error(self):
        return self.c_status < 0

    def __str__(self):
        return 'ApplicationReturnStatus(status={})'.format(self.c_status)

    def __repr__(self):
        return '<{} at {}>'.format(str(self), hex(id(self)))



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pypopt is a wrapper around Ipopt C++ interface.
"""

__all__ = [
    'IpoptApplication', 'TNLP', 'TNLP', 'PythonJournal', 'EJournalLevel', 'EJournalCategory',
    'IndexStyle', 'NLPInfo',
]

# pylint: disable=no-name-in-module
from pypopt._cython import (
    IpoptApplication,
    TNLP,
    PythonJournal,
    EJournalLevel,
    EJournalCategory,
    IndexStyle,
    NLPInfo,
    SolverReturn,
)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file
import os
from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize
from pathlib import Path

project_root = Path(__file__).resolve().parent

about = {}
version_path = project_root / 'pypopt' / '__version__.py'
with version_path.open() as f:
    exec(f.read(), about)

with (project_root / 'README.rst').open() as f:
    readme = f.read()

with (project_root / 'CHANGELOG.rst').open() as f:
    changelog = f.read()

include_dirs = []
if os.environ.get('IPOPT_INCLUDE_DIR'):
    include_dirs.append(os.environ.get('IPOPT_INCLUDE_DIR'))

library_dirs = []
if os.environ.get('IPOPT_LIBRARY_DIR'):
    library_dirs.append(os.environ.get('IPOPT_LIBRARY_DIR'))

extensions = [Extension(
    'pypopt._cython',
    sources=[
        'pypopt/_cython.pyx',
    ],
    libraries=['ipopt'],
    language='c++',
    extra_compile_args=['-std=c++11'],
    extra_link_args=['-std=c++11'],
    include_dirs=include_dirs,
    library_dirs=library_dirs,
)]


setup(
    name='pypopt',
    description=about['__description__'],
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],
    long_description=readme + '\n\n' + changelog,
    packages=find_packages(exclude=['tests']),
    ext_modules=cythonize(extensions),
    include_package_data=True,
    install_requires=['numpy'],
    setup_requires=['pytest-runner', 'cython'],
    tests_require=['pytest', 'pytest-cov'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
)



from cython.operator cimport dereference as d
from libcpp.string cimport string
cimport ipopt as ip


cdef class OptionsList:
    cdef ip.SmartPtr[ip.OptionsList] c_options

    def clear(self):
        return d(self.c_options).clear()

    def get_list(self):
        cdef string c_outlist
        d(self.c_options).PrintList(c_outlist)
        # convert to python str
        outlist = c_outlist.decode('utf-8')
        lines = outlist.split('\n')
        opts = []
        # skip first (header) and last (empty)
        for line in lines[1:-1]:
            o = line.split()
            opts.append({
                'name': o[0],
                'value': o[2],
                'times_used': o[3],
            })
        return opts

    def set_string_value(self, str tag, str value, allow_clobber=True, dont_print=False):
        cdef string c_tag = tag.encode('utf-8')
        cdef string c_value = value.encode('utf-8')
        return d(self.c_options).SetStringValue(c_tag, c_value, allow_clobber, dont_print)

    def set_numeric_value(self, str tag, ip.Number value, allow_clobber=True, dont_print=False):
        cdef string c_tag = tag.encode('utf-8')
        return d(self.c_options).SetNumericValue(c_tag, value, allow_clobber, dont_print)

    def set_integer_value(self, str tag, ip.Index value, allow_clobber=True, dont_print=False):
        cdef string c_tag = tag.encode('utf-8')
        return d(self.c_options).SetIntegerValue(c_tag, value, allow_clobber, dont_print)

    def set_string_value_if_unset(self, str tag, str value, allow_clobber=True, dont_print=False):
        cdef string c_tag = tag.encode('utf-8')
        cdef string c_value = value.encode('utf-8')
        return d(self.c_options).SetStringValueIfUnset(c_tag, c_value, allow_clobber, dont_print)

    def set_numeric_value_if_unset(self, str tag, ip.Number value, allow_clobber=True, dont_print=False):
        cdef string c_tag = tag.encode('utf-8')
        return d(self.c_options).SetNumericValueIfUnset(c_tag, value, allow_clobber, dont_print)

    def set_integer_value_if_unset(self, str tag, ip.Index value, allow_clobber=True, dont_print=False):
        cdef string c_tag = tag.encode('utf-8')
        return d(self.c_options).SetIntegerValueIfUnset(c_tag, value, allow_clobber, dont_print)

    def get_string_value(self, str tag, str prefix=''):
        cdef string c_tag = tag.encode('utf-8')
        cdef string c_prefix = tag.encode('utf-8')
        cdef string c_value
        found = d(self.c_options).GetStringValue(c_tag, c_value, c_prefix)
        if not found:
            return
        return c_value.decode('utf-8')

    def get_enum_value(self, str tag, str prefix=''):
        cdef string c_tag = tag.encode('utf-8')
        cdef string c_prefix = tag.encode('utf-8')
        cdef ip.Index c_value = -1
        found = d(self.c_options).GetEnumValue(c_tag, c_value, c_prefix)
        if not found:
            return
        return c_value

    def get_bool_value(self, str tag, str prefix=''):
        cdef string c_tag = tag.encode('utf-8')
        cdef string c_prefix = tag.encode('utf-8')
        cdef bool c_value = False
        found = d(self.c_options).GetBoolValue(c_tag, c_value, c_prefix)
        if not found:
            return
        return c_value

    def get_numeric_value(self, str tag, str prefix=''):
        cdef string c_tag = tag.encode('utf-8')
        cdef string c_prefix = tag.encode('utf-8')
        cdef ip.Number c_value = -1
        found = d(self.c_options).GetNumericValue(c_tag, c_value, c_prefix)
        if not found:
            return
        return c_value

    def get_integer_value(self, str tag, str prefix=''):
        cdef string c_tag = tag.encode('utf-8')
        cdef string c_prefix = tag.encode('utf-8')
        cdef ip.Index c_value = -1
        found = d(self.c_options).GetIntegerValue(c_tag, c_value, c_prefix)
        if not found:
            return
        return c_value



from pypopt.coin.typedef cimport *
from pypopt.coin.alg_types cimport *
from pypopt.coin.smart_ptr cimport *
from pypopt.coin.journalist cimport *
from pypopt.coin.return_codes cimport *
from pypopt.coin.options_list cimport *
from pypopt.coin.ipopt_application cimport *
from pypopt.coin.solve_statistics cimport *
from pypopt.coin.tnlp cimport *
